# Copyright (c) 2016 NEC Corporation. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from aodhclient import client as aodh_client
from oslo_log import log as logging
import six

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)


class AodhDriver(datasource_driver.PollingDataSourceDriver,
                 datasource_driver.ExecutionDriver):
    ALARMS = "alarms"

    value_trans = {'translation-type': 'VALUE'}
    # TODO(ramineni): enable ALARM_RULES translator

    alarms_translator = {
        'translation-type': 'HDICT',
        'table-name': ALARMS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'alarm_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'enabled', 'translator': value_trans},
             {'fieldname': 'type', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'time_constraints', 'translator': value_trans},
             {'fieldname': 'user_id', 'translator': value_trans},
             {'fieldname': 'project_id', 'translator': value_trans},
             {'fieldname': 'alarm_actions', 'translator': value_trans},
             {'fieldname': 'ok_actions', 'translator': value_trans},
             {'fieldname': 'insufficient_data_actions', 'translator':
             value_trans},
             {'fieldname': 'repeat_actions', 'translator': value_trans},
             {'fieldname': 'timestamp', 'translator': value_trans},
             {'fieldname': 'state_timestamp', 'translator': value_trans},
             )}

    def safe_id(x):
        if isinstance(x, six.string_types):
            return x
        try:
            return x['resource_id']
        except KeyError:
            return str(x)

    TRANSLATORS = [alarms_translator]

    def __init__(self, name='', args=None):
        super(AodhDriver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        session = ds_utils.get_keystone_session(args)
        endpoint = session.get_endpoint(service_type='alarming',
                                        interface='publicURL')
        self.aodh_client = aodh_client.Client(version='2', session=session,
                                              endpoint_override=endpoint)
        self.add_executable_client_methods(self.aodh_client, 'aodhclient.v2.')
        self.initialize_update_method()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'aodh'
        result['description'] = ('Datasource driver that interfaces with '
                                 'aodh.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_method(self):
        alarms_method = lambda: self._translate_alarms(
            self.aodh_client.alarm.list())
        self.add_update_method(alarms_method, self.alarms_translator)

    @ds_utils.update_state_on_changed(ALARMS)
    def _translate_alarms(self, obj):
        """Translate the alarms represented by OBJ into tables."""
        LOG.debug("ALARMS: %s", str(obj))
        row_data = AodhDriver.convert_objs(obj, self.alarms_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.aodh_client, action, action_args)



# Copyright (c) 2015 Cisco, 2018 NEC, Inc.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from datetime import datetime
from datetime import timedelta

import eventlet
from futurist import periodics
from monascaclient import client as monasca_client
from oslo_concurrency import lockutils
from oslo_log import log as logging

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils

LOG = logging.getLogger(__name__)

DATA = "statistics.data"
DIMENSIONS = "dimensions"
METRICS = "metrics"
NOTIFICATIONS = "alarm_notification"
STATISTICS = "statistics"
value_trans = {'translation-type': 'VALUE'}


# TODO(thinrichs): figure out how to move even more of this boilerplate
#   into DataSourceDriver.  E.g. change all the classes to Driver instead of
#   NeutronDriver, CeilometerDriver, etc. and move the d6instantiate function
#   to DataSourceDriver.
class MonascaDriver(datasource_driver.PollingDataSourceDriver,
                    datasource_driver.ExecutionDriver):

    # TODO(fabiog): add events and logs when fully supported in Monasca
    # EVENTS = "events"
    # LOGS = "logs"

    metric_translator = {
        'translation-type': 'HDICT',
        'table-name': METRICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'dimensions',
              'translator': {'translation-type': 'VDICT',
                             'table-name': DIMENSIONS,
                             'id-col': 'id',
                             'key-col': 'key', 'val-col': 'value',
                             'translator': value_trans}})
    }

    statistics_translator = {
        'translation-type': 'HDICT',
        'table-name': STATISTICS,
        'selector-type': 'DICT_SELECTOR',
        'field-translators':
            ({'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'statistics',
              'translator': {'translation-type': 'LIST',
                             'table-name': DATA,
                             'id-col': 'name',
                             'val-col': 'value_col',
                             'translator': value_trans}})
    }

    TRANSLATORS = [metric_translator, statistics_translator]

    def __init__(self,  name='', args=None):
        super(MonascaDriver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        if not args.get('project_name'):
            args['project_name'] = args['tenant_name']

        # set default polling time to 1hr
        self.poll_time = int(args.get('poll_time', 3600))

        session = ds_utils.get_keystone_session(args)

        # if the endpoint not defined retrieved it from keystone catalog
        if 'endpoint' not in args:
            args['endpoint'] = session.get_endpoint(service_type='monitoring',
                                                    interface='publicURL')

        self.monasca = monasca_client.Client('2_0', session=session,
                                             endpoint=args['endpoint'])
        self.add_executable_client_methods(self.monasca, 'monascaclient.')
        self.initialize_update_methods()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'monasca'
        result['description'] = ('Datasource driver that interfaces with '
                                 'monasca.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_methods(self):
        metrics_method = lambda: self._translate_metric(
            self.monasca.metrics.list())
        self.add_update_method(metrics_method, self.metric_translator)

        statistics_method = self.update_statistics
        self.add_update_method(statistics_method, self.statistics_translator)

    def update_statistics(self):
        today = datetime.utcnow()
        yesterday = timedelta(hours=24)
        start_from = datetime.isoformat(today-yesterday)

        for metric in self.monasca.metrics.list_names():
            LOG.debug("Monasca statistics for metric %s", metric['name'])
            _query_args = dict(
                start_time=start_from,
                name=metric['name'],
                statistics='avg',
                period=int(self.poll_time),
                merge_metrics='true')
            statistics = self.monasca.metrics.list_statistics(
                **_query_args)
            self._translate_statistics(statistics)

    @ds_utils.update_state_on_changed(METRICS)
    def _translate_metric(self, obj):
        """Translate the metrics represented by OBJ into tables."""
        LOG.debug("METRIC: %s", str(obj))

        row_data = MonascaDriver.convert_objs(obj,
                                              self.metric_translator)
        return row_data

    @ds_utils.update_state_on_changed(STATISTICS)
    def _translate_statistics(self, obj):
        """Translate the metrics represented by OBJ into tables."""

        LOG.debug("STATISTICS: %s", str(obj))

        row_data = MonascaDriver.convert_objs(obj,
                                              self.statistics_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.monasca, action, action_args)


class MonascaWebhookDriver(datasource_driver.PushedDataSourceDriver):

    def flatten_alarm_webhook(alarm_obj):
        flattened = []
        key_to_sub_dict = 'metrics'
        for alarm in alarm_obj:
            sub_dict = alarm.pop(key_to_sub_dict)[0]
            for k, v in sub_dict.items():
                if isinstance(v, dict):
                    for key, value in v.items():
                        alarm[key_to_sub_dict + '_' + k + '_' + key] = value
                else:
                    alarm[key_to_sub_dict + '_' + k] = v
            flattened.append(alarm)
        return flattened

    alarm_notification_translator = {
        'translation-type': 'HDICT',
        'table-name': NOTIFICATIONS,
        'selector-type': 'DICT_SELECTOR',
        'objects-extract-fn': flatten_alarm_webhook,
        'field-translators':
            ({'fieldname': 'alarm_id', 'translator': value_trans},
             {'fieldname': 'alarm_definition_id', 'translator': value_trans},
             {'fieldname': 'alarm_name', 'translator': value_trans},
             {'fieldname': 'alarm_description', 'translator': value_trans},
             {'fieldname': 'alarm_timestamp', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'old_state', 'translator': value_trans},
             {'fieldname': 'message', 'translator': value_trans},
             {'fieldname': 'tenant_id', 'translator': value_trans},
             {'fieldname': 'metrics_id', 'col': 'first_metric_id',
              'translator': value_trans},
             {'fieldname': 'metrics_name', 'col': 'first_metric_name',
              'translator': value_trans},
             {'fieldname': 'metrics_dimensions_hostname',
              'col': 'first_metric_hostname',
              'translator': value_trans},
             {'fieldname': 'metrics_dimensions_service',
              'col': 'first_metric_service',
              'translator': value_trans},)
    }
    TRANSLATORS = [alarm_notification_translator]

    def __init__(self, name='', args=None):
        LOG.warning(
            'The Monasca webhook driver is classified as having unstable '
            'schema. The schema may change in future releases in '
            'backwards-incompatible ways.')
        super(MonascaWebhookDriver, self).__init__(name, args=args)
        if args is None:
            args = {}
        # set default time to 10 days before deleting an active alarm
        self.hours_to_keep_alarm = int(args.get('hours_to_keep_alarm', 240))
        self.set_up_periodic_tasks()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'monasca_webhook'
        result['description'] = ('Datasource driver that accepts Monasca '
                                 'webhook alarm notifications.')
        result['config'] = {'persist_data': constants.OPTIONAL,
                            'hours_to_keep_alarm': constants.OPTIONAL}
        return result

    def _webhook_handler(self, alarm):
        tablename = NOTIFICATIONS
        # remove already existing same alarm row
        alarm_id = alarm['alarm_id']
        column_index_number_of_alarm_id = 0
        to_remove = [row for row in self.state[tablename]
                     if row[column_index_number_of_alarm_id] == alarm_id]
        for row in to_remove:
            self.state[tablename].discard(row)

        translator = self.alarm_notification_translator
        row_data = MonascaWebhookDriver.convert_objs([alarm], translator)

        # add alarm to table
        for table, row in row_data:
            if table == tablename:
                self.state[tablename].add(row)
        LOG.debug('publish a new state %s in %s',
                  self.state[tablename], tablename)
        self.publish(tablename, self.state[tablename])
        return [tablename]

    def set_up_periodic_tasks(self):
        @lockutils.synchronized('congress_monasca_webhook_ds_data')
        @periodics.periodic(spacing=max(self.hours_to_keep_alarm * 3600/10, 1))
        def delete_old_alarms():
            tablename = NOTIFICATIONS
            col_index_of_timestamp = 4
            # find for removal all alarms at least self.hours_to_keep_alarm old
            to_remove = [
                row for row in self.state[tablename]
                if (datetime.utcnow() -
                    datetime.utcfromtimestamp(row[col_index_of_timestamp])
                    >= timedelta(hours=self.hours_to_keep_alarm))]
            for row in to_remove:
                self.state[tablename].discard(row)

        periodic_task_callables = [
            (delete_old_alarms, None, {}),
            (delete_old_alarms, None, {})]
        self.periodic_tasks = periodics.PeriodicWorker(periodic_task_callables)
        self.periodic_tasks_thread = eventlet.spawn_n(
            self.periodic_tasks.start)

    def __del__(self):
        if self.periodic_tasks:
            self.periodic_tasks.stop()
            self.periodic_tasks.wait()
            self.periodic_tasks = None
        if self.periodic_tasks_thread:
            eventlet.greenthread.kill(self.periodic_tasks_thread)
            self.periodic_tasks_thread = None



# Copyright (c) 2014 VMware, Inc. All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

from congress.api import api_utils
from congress.api import base
from congress.api import webservice
from congress import exception


class StatusModel(base.APIModel):
    """Model for handling API requests about Statuses."""

    # Note(thread-safety): blocking function
    def get_item(self, id_, params, context=None):
        """Retrieve item with id id\_ from model.

        :param: id\_: The ID of the item to retrieve
        :param: params: A dict-like object containing parameters
                    from the request query string and body.
        :param: context: Key-values providing frame of reference of request

        :returns: The matching item or None if item with id\_ does not exist.
        """
        # Note(thread-safety): blocking call
        caller, source_id = api_utils.get_id_from_context(context)
        # FIXME(threod-safety): in DSE2, the returned caller can be a
        #   datasource name. But the datasource name may now refer to a new,
        #   unrelated datasource. Causing the rest of this code to operate on
        #   an unintended datasource.
        #   Fix: check UUID of datasource before operating. Abort if mismatch

        try:
            rpc_args = {'params': context, 'source_id': source_id}
            # Note(thread-safety): blocking call
            status = self.invoke_rpc(caller, 'get_status', rpc_args)
        except exception.CongressException as e:
            raise webservice.DataModelException(
                exception.NotFound.code, str(e),
                http_status_code=exception.NotFound.code)

        return status



# Copyright 2017 VMware Inc. All rights reserved.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import copy

from congress import exception
from congress.library_service import library_service
from congress.tests import base


class TestLibraryService(base.SqlTestCase):

    def setUp(self):
        super(TestLibraryService, self).setUp()
        self.library = library_service.LibraryService('lib-test')
        self.library.delete_all_policies()  # clear pre-loaded library policies

        self.policy1 = {'name': 'policy1', 'abbreviation': 'abbr',
                        'kind': 'database', 'description': 'descrip',
                        'rules': [{'rule': 'p(x) :- q(x)',
                                   'comment': 'test comment',
                                   'name': 'testname'}]}

        self.policy2 = {'name': 'policy2', 'abbreviation': 'abbr',
                        'kind': 'database', 'description': 'descrip',
                        'rules': [{'rule': 'p(x) :- q(x)',
                                   'comment': 'test comment',
                                   'name': 'testname'}]}

        self.policy1_meta = copy.deepcopy(self.policy1)
        self.policy2_meta = copy.deepcopy(self.policy2)
        del self.policy1_meta['rules']
        del self.policy2_meta['rules']

    def test_create_policy_no_name(self):
        self.assertRaises(exception.InvalidPolicyInput,
                          self.library.create_policy, {'rules': []})

    def test_create_policy_no_rules(self):
        self.assertRaises(exception.InvalidPolicyInput,
                          self.library.create_policy, {'name': 'policy1'})

    def test_create_policy_other_schema_violations(self):
        # name too long (255 limit)
        policy_item = {
            'name': 'policy2', 'abbreviation': 'abbr',
            'kind': 'database', 'description': 'descrip',
            'rules': [{
                'rule': 'p(x) :- q(x)',
                'comment': 'test comment',
                'name':
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '11111111111111111111111111111'}]}
        self.assertRaises(exception.InvalidPolicyInput,
                          self.library.create_policy, policy_item)

        # comment too long (255 limit)
        policy_item = {
            'name': 'policy2', 'abbreviation': 'abbr',
            'kind': 'database', 'description': 'descrip',
            'rules': [{
                'rule': 'p(x) :- q(x)',
                'comment':
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '111111111111111111111111111111111111111111111111111111111'
                    '11111111111111111111111111111',
                'name': 'testname'}]}
        self.assertRaises(exception.InvalidPolicyInput,
                          self.library.create_policy, policy_item)

        # rule item missing 'rule' property
        policy_item = {
            'name': 'policy2', 'abbreviation': 'abbr',
            'kind': 'database', 'description': 'descrip',
            'rules': [{
                'comment': 'test comment',
                'name': 'testname'}]}
        self.assertRaises(exception.InvalidPolicyInput,
                          self.library.create_policy, policy_item)

    def test_create_policy_bad_name(self):
        self.assertRaises(exception.PolicyException,
                          self.library.create_policy,
                          {'name': 'disallowed-hyphen', 'rules': []})

    def test_create_policy_default(self):
        res = self.library.create_policy({'name': 'policy1', 'rules': []})
        self.assertEqual(res, {'id': res['id'], 'abbreviation': 'polic',
                               'kind': 'nonrecursive', 'name': 'policy1',
                               'description': '', 'rules': []})

    def test_create_policy(self):
        policy_obj = self.library.create_policy(self.policy1)
        self.policy1['id'] = policy_obj['id']
        self.assertEqual(policy_obj, self.policy1)

    def test_create_policy_duplicate(self):
        self.library.create_policy({'name': 'policy1', 'rules': []})
        self.assertRaises(KeyError, self.library.create_policy,
                          {'name': 'policy1', 'rules': []})
        res = self.library.get_policies()
        self.assertEqual(len(res), 1)

    def test_get_policy_empty(self):
        res = self.library.get_policies()
        self.assertEqual(res, [])

        self.assertRaises(KeyError, self.library.get_policy,
                          'nosuchpolicy')

        self.assertRaises(KeyError, self.library.get_policy_by_name,
                          'nosuchpolicy')

    def test_create_get_policy(self):
        policy_obj = self.library.create_policy(self.policy1)
        self.policy1['id'] = policy_obj['id']
        self.policy1_meta['id'] = policy_obj['id']
        res = self.library.get_policies()
        self.assertEqual(res, [self.policy1])

        res = self.library.get_policy(policy_obj['id'])
        self.assertEqual(res, self.policy1)

        res = self.library.get_policy_by_name(policy_obj['name'])
        self.assertEqual(res, self.policy1)

        res = self.library.get_policies(include_rules=True)
        self.assertEqual(res, [self.policy1])

        res = self.library.get_policy(policy_obj['id'], include_rules=False)
        self.assertEqual(res, self.policy1_meta)

        res = self.library.get_policy_by_name(policy_obj['name'],
                                              include_rules=False)
        self.assertEqual(res, self.policy1_meta)

        self.assertRaises(KeyError, self.library.get_policy, 'no_such_policy')

        self.assertRaises(KeyError, self.library.get_policy_by_name,
                          'no_such_policy')

    def test_delete_policy(self):
        self.assertRaises(KeyError, self.library.delete_policy,
                          'policy1')

        policy_obj = self.library.create_policy(self.policy1)
        self.policy1['id'] = policy_obj['id']

        policy_obj = self.library.create_policy(self.policy2)
        self.policy2['id'] = policy_obj['id']

        res = self.library.get_policies()
        self.assertEqual(len(res), 2)
        self.assertTrue(all(p in res
                            for p in [self.policy1, self.policy2]))

        self.assertRaises(KeyError, self.library.delete_policy,
                          'no_such_policy')

        res = self.library.delete_policy(self.policy1['id'])
        self.assertEqual(res, self.policy1)

        res = self.library.get_policies()
        self.assertEqual(len(res), 1)
        self.assertEqual(res[0], self.policy2)

        res = self.library.delete_policy(self.policy2['id'])
        self.assertEqual(res, self.policy2)

        res = self.library.get_policies()
        self.assertEqual(len(res), 0)

    def test_delete_policies(self):
        self.library.delete_all_policies()
        res = self.library.get_policies()
        self.assertEqual(len(res), 0)

        self.library.create_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        self.library.create_policy(
            {'name': 'policy2', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        self.library.delete_all_policies()
        res = self.library.get_policies()
        self.assertEqual(len(res), 0)

    def test_replace_policy(self):
        policy1 = self.library.create_policy(
            {'name': 'policy1', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        policy2 = self.library.create_policy(
            {'name': 'policy2', 'abbreviation': 'abbr', 'kind': 'database',
             'description': 'descrip', 'rules': [{'rule': 'p(x) :- q(x)',
                                                  'comment': 'test comment',
                                                  'name': 'testname'}]})

        replacement_policy = {
            "name": "new_name",
            "description": "new test policy2 description",
            "kind": "nonrecursive",
            "abbreviation": "newab",
            "rules": [{"rule": "r(x) :- c(x)", "comment": "test comment",
                       "name": "test name"}]
        }

        # update non-existent item
        self.assertRaises(KeyError,
                          self.library.replace_policy, 'no_such_id',
                          replacement_policy)

        # update existing item
        self.library.replace_policy(policy2['id'], replacement_policy)

        replacement_policy_w_id = copy.deepcopy(replacement_policy)
        replacement_policy_w_id['id'] = policy2['id']

        ret = self.library.get_policies()
        self.assertEqual(len(ret), 2)
        self.assertTrue(all(p in ret
                            for p in [policy1,
                                      replacement_policy_w_id]))



#
# Copyright (c) 2017 Orange.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Tests for the unmarshaling of options by the driver"""

from oslo_config import cfg
from oslo_config import types
from oslo_log import log as logging

from congress.cfg_validator import parsing
from congress.tests import base

LOG = logging.getLogger(__name__)


OPT_TEST = {
    u'positional': False, u'kind': u'BoolOpt',
    u'deprecated_reason': None,
    u'help': u'Enables or disables inter-process locks.',
    u'default': False, u'type': {u'type': u'Boolean'},
    u'required': False, u'sample_default': None,
    u'deprecated_opts': [{u'group': u'DEFAULT', u'name': None}],
    u'deprecated_for_removal': False,
    u'dest': u'disable_process_locking',
    u'secret': False, u'short': None, u'mutable': False,
    u'deprecated_since': None, u'metavar': None,
    u'advanced': False, u'name': u'disable_process_locking'}
DICT_NS_TEST = {
    u'DEFAULT': {u'object': None, u'namespaces': []},
    u'oslo_concurrency': {
        u'object': None,
        u'namespaces': [[u'oslo.concurrency', [OPT_TEST]]]}}


class TestParsing(base.TestCase):
    """Tests for the unmarshaling of options by the driver"""

    def test_add_namespace(self):
        """Test for adding a namespace"""
        conf = cfg.ConfigOpts()
        initial_keys_len = len(conf.keys())
        parsing.add_namespace(conf, DICT_NS_TEST, 'abcde-12345')
        keys = conf.keys()
        self.assertEqual(initial_keys_len + 1, len(keys))
        self.assertIn(u'oslo_concurrency', keys)
        self.assertIsNotNone(
            conf.get(u'oslo_concurrency').get(u'disable_process_locking'))

    def test_construct_conf_manager(self):
        """Test for building a conf manager"""
        initial_keys_len = len(cfg.ConfigOpts().keys())
        conf = parsing.construct_conf_manager([DICT_NS_TEST])
        self.assertIsInstance(conf, cfg.ConfigOpts)
        keys = conf.keys()
        self.assertEqual(initial_keys_len + 1, len(keys))
        self.assertIn(u'oslo_concurrency', keys)

    def test_make_group(self):
        """Test for parsing a group"""
        grp = parsing.make_group('group', 'group_title', 'group help')
        self.assertIsInstance(grp, cfg.OptGroup)
        self.assertEqual("group", grp.name)
        self.assertEqual("group_title", grp.title)

    def test_make_opt(self):
        """Test for parsing an option"""
        descr = {
            u'positional': False,
            u'kind': u'Opt',
            u'deprecated_reason': None,
            u'help': u'Help me',
            u'default': None,
            u'type': {u'type': u'String'},
            u'required': False, u'sample_default': None,
            u'deprecated_opts': [], u'deprecated_for_removal': False,
            u'dest': u'name',
            u'secret': False,
            u'short': None,
            u'mutable': False,
            u'deprecated_since': None,
            u'metavar': None,
            u'advanced': False,
            u'name': u'name'}
        opt = parsing.make_opt(descr, 'abcd-1234', 'efgh-5678')
        self.assertIsInstance(opt, parsing.IdentifiedOpt)
        self.assertEqual("name", opt.name)
        self.assertEqual('abcd-1234', opt.id_)
        self.assertEqual('efgh-5678', opt.ns_id)

    def test_make_type(self):
        """Test for parsing a type"""
        typ1 = parsing.make_type({u'type': u'String'})
        self.assertIsInstance(typ1, types.String)
        typ2 = parsing.make_type({u'type': u'Integer'})
        self.assertIsInstance(typ2, types.Integer)
        typ3 = parsing.make_type(
            {u'item_type': {u'type': u'Boolean'}, u'type': u'List'})
        self.assertIsInstance(typ3, types.List)
        self.assertIsInstance(typ3.item_type, types.Boolean)



#    Copyright 2018 Orange
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Unit tests for z3theory"""
import mock
import six

from congress import data_types
from congress.datalog import compile as ast
from congress.datalog import topdown
from congress import exception
from congress.tests import base
from congress.tests.z3 import z3mock as z3
from congress.z3 import z3theory
from congress.z3 import z3types


def mockz3(f):
    z3types.Z3_AVAILABLE = True
    return (
        mock.patch("congress.z3.z3types.z3", new=z3)
        (mock.patch("congress.z3.z3theory.Z3OPT", new=z3)(f)))


class TestZ3Utilities(base.TestCase):

    def test_cycle_not_contained_in_z3(self):
        t1 = mock.MagicMock(spec=z3theory.Z3Theory)
        t2 = mock.MagicMock(spec=z3theory.Z3Theory)
        t3 = mock.MagicMock(spec=topdown.TopDownTheory)
        theories = {'t1': t1, 't2': t2, 't3': t3}
        for name, th in six.iteritems(theories):
            th.name = name
        cycles = [['t1:p', 't2:q', 't1:r'], ['t1:p1', 't2:q2']]
        r = z3theory.cycle_not_contained_in_z3(theories, cycles)
        self.assertIs(False, r)
        cycles = [['t1:p', 't2:q', 't1:r'], ['t3:p1', 't2:q2']]
        r = z3theory.cycle_not_contained_in_z3(theories, cycles)
        self.assertIs(True, r)

    def test_congress_constant(self):
        test_cases = [
            (3, "INTEGER", 3), ("aa", "STRING", "aa"),
            (4.3, "FLOAT", 4.3), ([], "STRING", "[]")]
        for (val, typ, name) in test_cases:
            obj = z3theory.congress_constant(val)
            self.assertIsInstance(
                obj, ast.ObjectConstant,
                msg=('not a constant for %s' % (str(val))))
            self.assertEqual(typ, obj.type)
            self.assertEqual(name, obj.name)

    def test_retrieve(self):
        theory = mock.MagicMock(spec=topdown.TopDownTheory)
        theory.name = 'test'
        theory.schema = mock.MagicMock(spec=ast.Schema)
        theory.schema.arity.return_value = 3
        z3theory.retrieve(theory, 'table')
        args = theory.select.call_args
        query = args[0][0]
        self.assertIsInstance(query, ast.Literal)
        self.assertEqual('table', query.table.table)
        self.assertEqual('test', query.table.service)
        self.assertEqual(3, len(query.arguments))
        self.assertIs(
            True,
            all(isinstance(arg, ast.Variable) for arg in query.arguments))


class TestZ3Theory(base.TestCase):

    @mockz3
    def setUp(self):
        world = {}
        self.theory = z3theory.Z3Theory('test', theories=world)
        world['test'] = self.theory    # invariant to maintain in agnostic
        super(TestZ3Theory, self).setUp()

    def test_init(self):
        self.assertIsInstance(self.theory.schema, ast.Schema)
        self.assertIsInstance(self.theory.z3context, z3theory.Z3Context)
        context = z3theory.Z3Context.get_context()
        self.assertIn('test', context.z3theories)
        self.assertEqual(self.theory, context.z3theories['test'])

    def test_select(self):
        context = z3theory.Z3Context.get_context()
        context.select = mock.MagicMock()
        lit = ast.Literal(ast.Tablename('t'), [])
        self.theory.select(lit)
        context.select.assert_called_once_with(self.theory, lit, True)

    def test_drop(self):
        self.theory.drop()
        context = z3theory.Z3Context.get_context()
        self.assertNotIn('test', context.z3theories)

    def test_arity(self):
        lit = ast.Literal(ast.Tablename('t'),
                          [ast.Variable('x'), ast.Variable('x')])
        self.theory.insert(lit)
        self.assertEqual(2, self.theory.arity('t'))


def mkc(name, nullable=False):
    return {'type': name, 'nullable': nullable}


class TestZ3Context(base.TestCase):

    @mockz3
    def test_registration(self):
        context = z3theory.Z3Context()
        theory = mock.MagicMock(z3theory.Z3Theory)
        name = 'test'
        world = {}
        theory.name = name
        theory.theories = world
        world['foo'] = theory
        context.register(theory)
        self.assertIn(name, context.z3theories)
        self.assertEqual(theory, context.z3theories[name])
        self.assertEqual(world, context.theories)
        context.drop(theory)
        self.assertNotIn(name, context.z3theories)

    @mockz3
    def test_get_context(self):
        self.assertIsInstance(
            z3theory.Z3Context.get_context(), z3theory.Z3Context)

    @mockz3
    def test_declare_table(self):
        """Test single table declaration

        Declare table declares the relation of a single table from its type
        """
        context = z3theory.Z3Context()
        name = 'test'
        tbname = 'table'
        world = {}
        theory = z3theory.Z3Theory(name, theories=world)
        theory.schema.map[tbname] = [mkc('Int'), mkc('Int')]
        world[name] = theory
        context.declare_table(theory, tbname)
        # Only the mock as a _relations element and pylint is confused
        rels = context.context._relations  # pylint: disable=E1101
        self.assertEqual(1, len(rels))
        self.assertEqual(name + ':' + tbname, rels[0].name())
        self.assertEqual(3, len(rels[0]._typs))

    @mockz3
    def test_declare_tables(self):
        """Test declaration of internal z3theories tables

        Declare tables must iterate over all schemas and create relations
        with the right arity and types
        """
        context = z3theory.Z3Context()
        world = {}
        t1 = z3theory.Z3Theory('t1', theories=world)
        t2 = z3theory.Z3Theory('t2', theories=world)
        world['t1'] = t1
        world['t2'] = t2
        t1.schema.map['p'] = [mkc('Int')]
        t1.schema.map['q'] = [mkc('Str'), mkc('Str')]
        t2.schema.map['k'] = [mkc('Bool')]
        context.register(t1)
        context.register(t2)
        context.declare_tables()
        # Only the mock as a _relations element and pylint is confused
        rels = context.context._relations  # pylint: disable=E1101
        self.assertEqual(3, len(rels))
        self.assertIn('t1:q', context.relations)
        self.assertEqual(2, len(context.relations['t1:p']._typs))

    def init_three_theories(self):
        context = z3theory.Z3Context()
        world = {}
        for name in ['t1', 't2', 't3']:
            world[name] = z3theory.Z3Theory(name, theories=world)
        t1, t2, t3 = world['t1'], world['t2'], world['t3']
        context.register(t1)
        context.register(t2)
        # t3 is kept external
        # Declare rules
        for rule in ast.parse('p(x) :- t2:r(x), t3:s(x). q(x) :- p(x). p(4).'):
            t1.insert(rule)
        for rule in ast.parse('r(x) :- t1:p(x).'):
            t2.insert(rule)
        # typechecker
        t1.schema.map['p'] = [mkc('Int')]
        t1.schema.map['q'] = [mkc('Int')]
        t2.schema.map['r'] = [mkc('Int')]
        t3.schema.map['s'] = [mkc('Int')]
        t3.schema.map['t'] = [mkc('Int')]
        return context

    @mockz3
    def test_declare_external_tables(self):
        """Test declaration of internal z3theories tables

        Declare tables must iterate over all schemas and create relations
        with the right arity and types
        """
        context = self.init_three_theories()
        context.declare_external_tables()
        # Only the mock as a _relations element and pylint is confused
        rels = context.context._relations  # pylint: disable=E1101
        self.assertEqual(1, len(rels))
        self.assertIn('t3:s', context.relations)

    @mockz3
    def test_compile_facts(self):
        context = z3theory.Z3Context()
        world = {}
        t1 = z3theory.Z3Theory('t1', theories=world)
        world['t1'] = t1
        context.register(t1)
        for rule in ast.parse('l(1,2). l(3,4). l(5,6).'):
            t1.insert(rule)
        t1.schema.map['l'] = [mkc('Int'), mkc('Int')]
        context.declare_tables()
        context.compile_facts(t1)
        rules = context.context.get_rules()
        self.assertEqual(3, len(rules))
        self.assertIs(True, all(r.decl().name() == 't1:l' for r in rules))
        self.assertEqual(
            [[1, 2], [3, 4], [5, 6]],
            [[c.as_long() for c in r.children()] for r in rules])

    def init_one_rule(self):
        context = z3theory.Z3Context()
        world = {}
        t1 = z3theory.Z3Theory('t1', theories=world)
        world['t1'] = t1
        context.register(t1)
        rule = ast.parse('p(x) :- l(x,y), l(3,x).')[0]
        t1.schema.map['l'] = [mkc('Int'), mkc('Int')]
        t1.schema.map['p'] = [mkc('Int')]
        context.declare_tables()
        return (context, t1, rule)

    def init_one_theory(self, prog):
        context = z3theory.Z3Context()
        world = {}
        t1 = z3theory.Z3Theory('t1', theories=world)
        world['t1'] = t1
        context.register(t1)
        for rule in ast.parse(prog):
            t1.insert(rule)
        return (context, t1)

    @mockz3
    def test_compile_atoms(self):
        (context, t1, rule) = self.init_one_rule()
        result = context.compile_atoms(t1, rule.head, rule.body)
        self.assertEqual(3, len(result))
        (vars, head, body) = result
        self.assertEqual(2, len(vars))
        # two variables
        self.assertIs(
            True, all(x.decl().kind() == z3.Z3_OP_UNINTERPRETED for x in vars))
        # two literals in the body
        self.assertEqual(2, len(body))
        # Head literal is p
        self.assertEqual('t1:p', head.decl().name())
        # First body literal is l
        self.assertEqual('t1:l', body[0].decl().name())
        # First arg of second body literal is a compiled int constant
        self.assertEqual(3, body[1].children()[0].as_long())
        # Second arg of second body literal is a variable
        self.assertEqual(z3.Z3_OP_UNINTERPRETED,
                         body[1].children()[1].decl().kind())

    @mockz3
    def test_compile_rule(self):
        (context, t1, rule) = self.init_one_rule()
        context.compile_rule(t1, rule)
        result = context.context._rules[0]   # pylint: disable=E1101
        self.assertEqual('forall', result.decl().name())
        self.assertEqual('=>', result.children()[2].decl().name())
        self.assertEqual(
            'and', result.children()[2].children()[0].decl().name())

    @mockz3
    def test_compile_query(self):
        (context, t1, rule) = self.init_one_rule()
        result = context.compile_query(t1, rule.head)
        self.assertEqual('exists', result.decl().name())
        self.assertEqual('t1:p', result.children()[1].decl().name())

    @mockz3
    def test_compile_theory(self):
        context = self.init_three_theories()
        context.declare_tables()
        context.declare_external_tables()
        context.compile_theory(context.z3theories['t1'])
        rules = context.context._rules     # pylint: disable=E1101
        self.assertEqual(3, len(rules))

    @mockz3
    def test_compile_all(self):
        context = self.init_three_theories()
        context.compile_all()
        rules = context.context._rules     # pylint: disable=E1101
        rels = context.context._relations  # pylint: disable=E1101
        self.assertEqual(4, len(rules))
        self.assertEqual(['t1:p', 't1:q', 't2:r', 't3:s'],
                         sorted([k.name() for k in rels]))

    @staticmethod
    def mk_z3_result(context):
        sort = context.type_registry.get_type('Int')

        def vec(x):
            return z3.BitVecVal(x, sort)

        x, y = z3.Const('x', sort), z3.Const('y', sort)
        return z3.Or(z3.And(z3.Eq(x, vec(1)), z3.Eq(y, vec(2))),
                     z3.And(z3.Eq(x, vec(3)), z3.Eq(y, vec(4))))

    @mock.patch('congress.z3.z3types.z3.Fixedpoint.get_answer')
    @mockz3
    def test_eval(self, mock_get_answer):
        (context, t1) = self.init_one_theory('l(1,2). l(3,4).')
        expr = self.mk_z3_result(context)
        mock_get_answer.return_value = expr
        query = ast.parse('l(x,y)')[0]
        result = context.eval(t1, query)
        self.assertEqual(2, len(result[1]))
        self.assertEqual(2, len(result[2]))
        self.assertIs(True, all(len(row) == 2 for row in result[0]))

    @mock.patch('congress.z3.z3types.z3.Fixedpoint.get_answer')
    @mockz3
    def test_select(self, mock_get_answer):
        (context, t1) = self.init_one_theory('l(1,2). l(3,4).')
        expr = self.mk_z3_result(context)
        mock_get_answer.return_value = expr
        query = ast.parse('l(x,y)')[0]
        result = context.select(t1, query, True)
        self.assertEqual(ast.parse('l(1,2). l(3,4).'), result)

    @mockz3
    def test_inject(self):
        theory = mock.MagicMock(spec=topdown.TopDownTheory)
        world = {'t': theory}
        theory.name = 't'
        theory.schema = ast.Schema()
        theory.schema.map['l'] = [mkc('Int'), mkc('Int')]
        theory.select.return_value = ast.parse('l(1,2). l(3,4). l(5,6)')
        context = z3theory.Z3Context()
        # An external theory world
        context.theories = world
        # inject the declaration of external relation without rules
        param_types = [
            context.type_registry.get_type(typ)
            for typ in ['Int', 'Int', 'Bool']]
        relation = z3.Function('t:l', *param_types)
        context.context.register_relation(relation)
        context.relations['t:l'] = relation
        # the test
        context.inject('t', 'l')
        rules = context.context._rules  # pylint: disable=E1101
        self.assertIs(True, all(r.decl().name() == 't:l' for r in rules))
        self.assertEqual(
            [[1, 2], [3, 4], [5, 6]],
            sorted([[c.as_long() for c in r.children()] for r in rules]))


class TestTypeConstraints(base.TestCase):

    @mockz3
    def setUp(self):
        try:
            data_types.TypesRegistry.type_class('Small')
        except KeyError:
            typ = data_types.create_congress_enum_type(
                'Small', ['a', 'b'], data_types.Str)
            data_types.TypesRegistry.register(typ)
        super(TestTypeConstraints, self).setUp()

    @staticmethod
    def init_two_theories(prog):
        context = z3theory.Z3Context()
        world = {}
        for name in ['t1', 't2']:
            world[name] = z3theory.Z3Theory(name, theories=world)
        t1, t2 = world['t1'], world['t2']
        context.register(t1)
        # t2 is kept external
        # Declare rules
        for rule in ast.parse(prog):
            t1.insert(rule)
        # typechecker
        t2.schema.map['p'] = [mkc('Small')]
        t2.schema.map['q'] = [mkc('Str')]
        return context

    @mockz3
    def test_compile_all_ok(self):
        context = self.init_two_theories('p("a"). p(x) :- t2:p(x).')
        context.typecheck()
        context.compile_all()
        rules = context.context._rules     # pylint: disable=E1101
        self.assertEqual(2, len(rules))

    @mockz3
    def test_compile_fails_constraints(self):
        # Two external tables with different types: caught by typechecker.
        context = self.init_two_theories('p(x) :- t2:p(x), t2:q(x).')
        self.assertRaises(exception.PolicyRuntimeException, context.typecheck)

    @mockz3
    def test_compile_fails_values(self):
        context = self.init_two_theories('p("c"). p(x) :- t2:p(x).')
        # Type-check succeeds
        context.typecheck()
        # But we are caught when trying to convert 'c' as a 'Small' value
        self.assertRaises(exception.PolicyRuntimeException,
                          context.compile_all)



# Copyright (c) 2014 Montavista Software, LLC.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
#

"""Schema version history

version: 2.1
date: 2016-03-27
changes:

 - Added columns to the volumes table: encrypted, availability_zone,
   replication_status, multiattach, snapshot_id, source_volid,
   consistencygroup_id, migration_status
 - Added the attachments table for volume attachment information.

version: 2.0
Initial schema version.
"""

from __future__ import print_function
from __future__ import division
from __future__ import absolute_import

import cinderclient.client

from congress.datasources import constants
from congress.datasources import datasource_driver
from congress.datasources import datasource_utils as ds_utils


class CinderDriver(datasource_driver.PollingDataSourceDriver,
                   datasource_driver.ExecutionDriver):
    VOLUMES = "volumes"
    ATTACHMENTS = "attachments"
    SNAPSHOTS = "snapshots"
    SERVICES = "services"

    # This is the most common per-value translator, so define it once here.
    value_trans = {'translation-type': 'VALUE'}

    volumes_translator = {
        'translation-type': 'HDICT',
        'table-name': VOLUMES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'size', 'translator': value_trans},
             {'fieldname': 'user_id', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'description', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'bootable', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans},
             {'fieldname': 'volume_type', 'translator': value_trans},
             {'fieldname': 'encrypted', 'translator': value_trans},
             {'fieldname': 'availability_zone', 'translator': value_trans},
             {'fieldname': 'replication_status', 'translator': value_trans},
             {'fieldname': 'multiattach', 'translator': value_trans},
             {'fieldname': 'snapshot_id', 'translator': value_trans},
             {'fieldname': 'source_volid', 'translator': value_trans},
             {'fieldname': 'consistencygroup_id', 'translator': value_trans},
             {'fieldname': 'migration_status', 'translator': value_trans},
             {'fieldname': 'attachments',
              'translator': {'translation-type': 'HDICT',
                             'table-name': ATTACHMENTS,
                             'parent-key': 'id',
                             'parent-col-name': 'volume_id',
                             'parent-key-desc': 'UUID of volume',
                             'selector-type': 'DICT_SELECTOR',
                             'in-list': True,
                             'field-translators':
                                 ({'fieldname': 'server_id',
                                   'translator': value_trans},
                                  {'fieldname': 'attachment_id',
                                   'translator': value_trans},
                                  {'fieldname': 'host_name',
                                   'translator': value_trans},
                                  {'fieldname': 'device',
                                   'translator': value_trans})}}
             )}

    snapshots_translator = {
        'translation-type': 'HDICT',
        'table-name': SNAPSHOTS,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'id', 'translator': value_trans},
             {'fieldname': 'size', 'translator': value_trans},
             {'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'volume_id', 'translator': value_trans},
             {'fieldname': 'name', 'translator': value_trans},
             {'fieldname': 'created_at', 'translator': value_trans})}

    services_translator = {
        'translation-type': 'HDICT',
        'table-name': SERVICES,
        'selector-type': 'DOT_SELECTOR',
        'field-translators':
            ({'fieldname': 'status', 'translator': value_trans},
             {'fieldname': 'binary', 'translator': value_trans},
             {'fieldname': 'zone', 'translator': value_trans},
             {'fieldname': 'state', 'translator': value_trans},
             {'fieldname': 'updated_at', 'translator': value_trans},
             {'fieldname': 'host', 'translator': value_trans},
             {'fieldname': 'disabled_reason', 'translator': value_trans})}

    TRANSLATORS = [volumes_translator, snapshots_translator,
                   services_translator]

    def __init__(self, name='', args=None):
        super(CinderDriver, self).__init__(name, args=args)
        datasource_driver.ExecutionDriver.__init__(self)
        session = ds_utils.get_keystone_session(args)
        self.cinder_client = cinderclient.client.Client(version='2',
                                                        session=session)
        self.add_executable_client_methods(self.cinder_client,
                                           'cinderclient.v2.')
        self.initialize_update_method()
        self._init_end_start_poll()

    @staticmethod
    def get_datasource_info():
        result = {}
        result['id'] = 'cinder'
        result['description'] = ('Datasource driver that interfaces with '
                                 'OpenStack cinder.')
        result['config'] = ds_utils.get_openstack_required_config()
        result['config']['lazy_tables'] = constants.OPTIONAL
        result['secret'] = ['password']
        return result

    def initialize_update_method(self):
        volumes_method = lambda: self._translate_volumes(
            self.cinder_client.volumes.list(detailed=True,
                                            search_opts={'all_tenants': 1}))
        self.add_update_method(volumes_method, self.volumes_translator)

        snapshots_method = lambda: self._translate_snapshots(
            self.cinder_client.volume_snapshots.list(
                detailed=True, search_opts={'all_tenants': 1}))
        self.add_update_method(snapshots_method, self.snapshots_translator)

        services_method = lambda: self._translate_services(
            self.cinder_client.services.list(host=None, binary=None))
        self.add_update_method(services_method, self.services_translator)

    @ds_utils.update_state_on_changed(VOLUMES)
    def _translate_volumes(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.volumes_translator)
        return row_data

    @ds_utils.update_state_on_changed(SNAPSHOTS)
    def _translate_snapshots(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.snapshots_translator)
        return row_data

    @ds_utils.update_state_on_changed(SERVICES)
    def _translate_services(self, obj):
        row_data = CinderDriver.convert_objs(obj, self.services_translator)
        return row_data

    def execute(self, action, action_args):
        """Overwrite ExecutionDriver.execute()."""
        # action can be written as a method or an API call.
        func = getattr(self, action, None)
        if func and self.is_executable(func):
            func(action_args)
        else:
            self._execute_api(self.cinder_client, action, action_args)
